/* dataset.h -*- C++ -*- dataset interface
 * Rob Earhart
 * $Id: dataset.h,v 1.12 1998/03/09 00:58:09 rob Exp $
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_DATASET_H
#define ACAP_DATASET_H

#include <vector.h>
#include "refcount.h"
#include "ps.h"
#include "entry.h"
#include "dstring.h"
#include "searchable.h"

// Datas are persistent store objects with dataset semantics.

// I'm not entirely pleased with the structure of this code.
// The object-on-disk / adapter-in-memory distinction seems a little
// forced; it might be better to consolidate the two (so, for instance,
// entryref_t would be replaced by Entry, with the appropriate
// protections and Entry methods).

class NotifyContext;

class Dataset : public Searchable, private PS {
public:
  Dataset(const String &path, int fd = -1);
  void Noref();
  void Init(Trx *trx) { PS::Init(trx); }
  Mutex &InitMutex() { return init_mutex; }

  typedef vector<Entry> dsentries_t;
  typedef hash_set<DString, hash<DString>, equal_to<DString> > names_t;

  P<PSResources> GetResources() { return PS::GetResources(); }

  void GetEntries(dsentries_t &entries,
		  Timestamp::Stamp stamp,
		  DatasetGroup *group);

  Entry FindEntry(const String &name,
		  Timestamp::Stamp stamp);

  // Note; InsertEntry will delete a supplied oldentry.  Since this
  // may move entries, any active entry pointers are instantly
  // invalidated.
  Entry InsertEntry(const String &name,
		    Entry *oldentry,
		    Trx *trx);

  void Search(entries_t &result,
	      DatasetGroup *group,
	      NotifyContext *context,
	      const String &path,
	      Modifiers *modifiers,
	      Criteria *criteria,
	      Referer *referer,
	      Timestamp::Stamp stamp);

  void DeletedSince(names_t &deleted,
		    DatasetGroup *group,
		    Timestamp::Stamp stamp,
		    Timestamp::Stamp synctime);

  static P<Dataset> TraversePath(const String &basepath,
				 Timestamp::Stamp stamp,
				 Trx *trx = NULL,
				 NotifyContext *context = NULL);

private:
  // These are friends so that they can access the underlying PS object
  // when storing data.
  friend class Entry;
  friend class Attribute;
  friend class Store;
  friend class DatasetMap;

  ~Dataset();
  void GetDeleted(names_t &deleted,
		  names_t &defaulted,
		  names_t &exist,
		  Timestamp::Stamp stamp,
		  Timestamp::Stamp synctime);

  // Dataset functions
  caddr_t CreateRoot(Trx *trx, off_t *offset);

  void SwizzleObject(caddr_t object, unsigned char tag);

  // Dataset data
  Mutex init_mutex;
  datasetobject_t *GetData() const { return (datasetobject_t *)Root(); }
};

class DatasetMap {
public:
  static DatasetMap *Instance();
  P<Dataset> Lookup(const String &path, Trx *trx = NULL);
  P<Dataset> CreateDataset(Entry entry,
			   Trx *trx); // sets up the entry's subdataset attribute...
private:
  friend class Dataset;
  void RemoveDataset(const String &path, Dataset *d);
  Mutex &DatasetsMutex() { return datasets_mutex; }

  static void Initialize();
  DatasetMap() {}
  static DatasetMap *instance;
  static pthread_once_t instance_init;

  typedef hash_map<String, Dataset *, hash<String>, equal_to<String> > datasets_t;
  datasets_t datasets;
  Mutex datasets_mutex;
};

#endif // ACAP_DATA_H
