/* timestamp.h -*- C++ -*- timestamp interface
 * Rob Earhart
 * $Id: timestamp.h,v 1.4 1998/03/09 00:58:55 rob Exp $
 */
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef ACAP_TIMESTAMP_H
#define ACAP_TIMESTAMP_H

#include <sys/types.h>
#include <pthread.h>
#include <time.h>
#include <hash_set.h>
#include <set.h>
#include "str.h"
#include "ptr.h"
#include "mutex.h"
#include "grab.h"
#include "void.h"

class PSBlob;

class Timestamp {
public:
  struct Stamp {
    Stamp() : sec(0), frac(0) {}
    Stamp(time_t _sec, unsigned long _frac) : sec(_sec), frac(_frac) {}

    // Constructs a timestamp from a long string form.
    Stamp(String &stamp);
    operator void *() const { return (void *)(sec || frac ? 1 : 0); }
    bool operator!() const { return ! (sec || frac); }
    bool operator<(const Stamp &s) const {
      return (sec < s.sec || sec == s.sec && frac < s.frac);
    }
    void Zero() { sec = 0; frac = 0; }

    time_t sec;
    unsigned long frac;

    friend ostream &operator<<(ostream &o, const Stamp &stamp) {
      return (o << '(' << stamp.sec << ", " << stamp.frac << ')');
    }
  };

  const char *LongStr() {	// full blown version timestamp for clients
    if (! longstr)
      longstr = MakeLongStr(stamp);
    return longstr;
  }
  const char *ShortStr() {	// short version (for on-disk)
    if (! shortstr)
      shortstr = MakeShortStr(stamp);
    return shortstr;
  }
  const Stamp &GetStamp() const { return stamp; }

  static APtr<char> ToLongString(const Stamp &stamp) {
    APtr<char> str = MakeLongStr(stamp);
    return str;
  }

  // Constructs the highest possible timestamp
  static Stamp MaxStamp() {
    return Stamp(LONG_MAX, ULONG_MAX);
  }

  void AddPSBlob(PSBlob *blob);

  ~Timestamp();

private:
  friend class TimestampFactory;

  Timestamp(time_t sec, unsigned long frac)
    : stamp(sec, frac) {}

  static APtr<char> MakeLongStr(const Stamp &stamp);
  static APtr<char> MakeShortStr(const Stamp &stamp);

  APtr<char> longstr;
  APtr<char> shortstr;
  Stamp stamp;

  typedef hash_set<P<PSBlob>, hash<void *>, equal_to<void *> > blobs_t;
  blobs_t blobs;
  Mutex blobs_mutex;
};

class TimestampFactory {
public:
  static TimestampFactory *Instance();
  Ptr<Timestamp> NewStamp();
  void RemoveStamp(Timestamp *t);
  Timestamp::Stamp OldestActiveStamp() {
    Grab g(factory_mutex);
    return stamps.empty() ? Timestamp::MaxStamp() : (*(stamps.begin()))->GetStamp();
  }
private:
  TimestampFactory() {}
  static void Initialize();
  static TimestampFactory *instance;
  static pthread_once_t instance_init;

  Mutex factory_mutex;
  // Note: stamps_t must be a refinement of ReversibleContainer (or, actually,
  // some container type whose iterators are bidirectional).  Set works nicely...
  typedef set<Timestamp *, less<Timestamp *> > stamps_t;
  stamps_t stamps;
  Timestamp::Stamp last;
};

#endif // ACAP_TIMESTAMP_H
